# Generated by Django 3.0.7 on 2020-06-25 15:08

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0039_auto_20200131_1758'),
        ('geostore', '0038_auto_20200625_1244'),
    ]

    operations = [
    ]
