# Generated by Django 2.0.8 on 2018-09-18 12:10

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0013_auto_20180919_1146'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='feature',
            name='from_date',
        ),
        migrations.RemoveField(
            model_name='feature',
            name='to_date',
        ),
    ]
