# Generated by Django 2.0.9 on 2018-11-14 08:06

try:
    from django.db.models import JSONField
except ImportError:  # TODO Remove when dropping Django releases < 3.1
    from django.contrib.postgres.fields import JSONField
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0016_auto_20181120_1059'),
    ]

    operations = [
        migrations.AddField(
            model_name='layer',
            name='settings',
            field=JSONField(blank=True, default=dict),
        ),
    ]
