# Generated by Django 3.0.2 on 2020-01-28 14:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0037_auto_20200127_1357'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feature',
            name='source',
            field=models.IntegerField(blank=True, editable=False, help_text='Internal field used by pgRouting', null=True),
        ),
        migrations.AlterField(
            model_name='feature',
            name='target',
            field=models.IntegerField(blank=True, editable=False, help_text='Internal field used by pgRouting', null=True),
        ),
    ]
